import urllib.parse

from bs4 import BeautifulSoup
import requests

from .PwdKeyAuth import PwdKeyAuth
from ..exceptions import InvalidToken, CarolApiResponseException


class PwdFluig(PwdKeyAuth):
    """

    Args:

        user: `str`
            Username fluig identity
        password: `str`
            Password
    """

    def __init__(self, user, password, lazy_login=False):
        self.user = user
        self.password = password
        self._token = None
        self.carol = None
        self.connector_id = None
        self.sess = requests.Session()
        self.lazy_login = lazy_login

    def _fluig_login(self, username, password):
        url = "https://totvs.fluigidentity.com/api/auth/v1/login"

        login_data = {
            'domain': 'totvs',
            'grant_type': 'password',
            'username': username,
            'password': password,
            'captchaReturnedValue': 'undefined',
        }
        headers = {
            'accept': 'application/json',
            'content-type': 'application/x-www-form-urlencoded',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896.127 Safari/537.36',
        }

        response = self.sess.request("POST", url, headers=headers,
                                     data=login_data, allow_redirects=True)

        if not response.ok:
            raise InvalidToken(response.text)
        return response

    def login(self, carol):
        """

        Args:

            carol: pycarol.Carol
                Carol() instance.

        Returns:
            None

        """
        self.carol = carol

        params = params = {
            "redirect": f"/{self.carol.domain}/carol-ui/carol-ui/",
            "org": f"{self.carol.organization}",
            "env": f"{self.carol.domain}",
        }
        headers = {}
        headers.update({'User-Agent': self.carol._user_agent})
        url = f'https://{self.carol.organization}.carol.ai/samlAuth/'

        resp = self.sess.get(url, headers=headers,
                             allow_redirects=True, params=params)

        soup = BeautifulSoup(resp.text, 'html.parser')
        tag = soup.find("form")
        if tag is None:
            raise Exception("Unable to login using fluig")

        tag = urllib.parse.unquote(
            tag['action'].replace('/cloudpass/?forward=%2F', ''))
        next_redirect = "https://totvs.fluigidentity.com/cloudpass/" + tag

        token = self._fluig_login(self.user, self.password)

        for i, j in token.json().items():
            self.sess.cookies[i] = str(j)

        resp = self.sess.get(
            next_redirect, headers=headers, allow_redirects=True)

        soup = BeautifulSoup(resp.text, 'html.parser')
        payload = {}
        for tag in soup.find_all("input", type="hidden", attrs={"name": ["RelayState", "SAMLResponse", "Signature", "SigAlg", "KeyInfo"]}):
            payload[tag['name']] = tag['value']

        url = f"https://{self.carol.organization}.carol.ai/api/v1/saml/ACS?orgSubdomain={self.carol.organization}"
        headers['content-type'] = 'application/x-www-form-urlencoded'
        carol_resp = self.sess.request(
            "POST", url, headers=headers, data=payload, allow_redirects=False)

        if not carol_resp.ok:
            raise CarolApiResponseException(carol_resp.text)
        self.access_token = carol_resp.headers['Location'].split(
            'handoff=')[-1]

        super().login(self.carol)



"""PyCarol - Connecting Carol to Python."""
import os
import tempfile

__version__ = "2.54.15"

__TEMP_STORAGE__ = os.path.join(tempfile.gettempdir(), "carolina/cache")

__CONNECTOR_PYCAROL__ = "f9953f6645f449baaccd16ab462f9b64"

_CAROL_METADATA_STAGING = [
    "mdmCounterForEntity",
    "mdmId",
    "mdmLastUpdated",
    "mdmCreated",
    "mdmConnectorId",
    "mdmDeleted",
]

_CAROL_METADATA_GOLDEN = [
    "mdmCounterForEntity",
    "mdmStagingCounter",
    "mdmId",
    "mdmCreated",
    "mdmLastUpdated",
    "mdmTenantId",
    "mdmEntityType",
    "mdmSourceEntityNames",
    "mdmCrosswalk",
    "mdmStagingRecord",
    "mdmApplicationIdMasterRecordId",
    "mdmPreviousIds",
    "mdmDeleted",
]

_CAROL_METADATA_UNTIE_GOLDEN = "mdmStagingCounter"
_CAROL_METADATA_UNTIE_STAGING = "mdmCounterForEntity"


_NEEDED_FOR_MERGE = ["mdmCounterForEntity", "mdmId", "mdmDeleted"]

_REJECTED_DM_COLS = [
    "mdmConnectorId",
    "mdmCreated",
    "mdmCreatedUser",
    "mdmCrosswalk",
    "mdmDetailFieldValues",
    "mdmEntityTemplateId",
    "mdmEntityType",
    "mdmErrors",
    "mdmId",
    "mdmLastUpdated",
    "mdmMasterFieldAndValues",
    "mdmStagingApplicationId",
    "mdmStagingEntityName",
    "mdmStagingRecord",
    "mdmTenantId",
    "mdmUpdatedUser",
]


from . import bigquery  # noqa
from .apps import Apps  # noqa
from .auth.PwdAuth import PwdAuth  # noqa
from .auth.ApiKeyAuth import ApiKeyAuth  # noqa
from .auth.PwdKeyAuth import PwdKeyAuth  # noqa
from .auth.PwdFluig import PwdFluig  # noqa
from .bigquery import BQ, BQStorage  # noqa
from .carol import Carol  # noqa
from .carol_api import CarolAPI  # noqa
from .carolina import Carolina  # noqa
from .cds import CDSGolden, CDSStaging  # noqa
from .connectors import Connectors  # noqa
from .data_models import DataModel  # noqa
from .logger import CarolHandler  # noqa
from .query import Query  # noqa
from .staging import Staging  # noqa
from .storage import Storage  # noqa
from .subscription import Subscription  # noqa
from .tasks import Tasks  # noqa



import copy
from pathlib import Path
import json
import os
import typing as T
import warnings

from dotenv import load_dotenv
import requests
from urllib3.util.retry import Retry

from . import __CONNECTOR_PYCAROL__, __version__
from . import exceptions
from .auth.ApiKeyAuth import ApiKeyAuth
from .auth.PwdAuth import PwdAuth
from .organization import Organization
from .tenant import Tenant


ResponseType = T.Union[requests.Response, T.Dict[str, T.Any], T.List]


class SwitchContext:
    def __init__(
        self,
        parent_context,
        env_name=None,
        env_id=None,
        org_id=None,
        org_name=None,
        app_name=None,
    ):
        self.parent_context = parent_context
        self.env_name = env_name
        self.env_id = env_id
        self.app_name = app_name
        self.org_id = org_id
        self.org_name = org_name

    def __enter__(self):
        self.parent_context.switch_environment(
            env_name=self.env_name,
            env_id=self.env_id,
            app_name=self.app_name,
            org_name=self.org_name,
            org_id=self.org_id,
        )
        return self.parent_context

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.parent_context


class Carol:

    """Handles all Carol`s API calls and log user in Carol.

    In case all parameters are `None`, pycarol will try yo find their values in the
    environment variables.
    The values are:

        1. `CAROLTENANT` for domain
        2. `CAROLAPPNAME` for app_name
        3. `CAROLAPPOAUTH` for auth
        4. `CAROLORGANIZATION` for organization
        5. `CAROLCONNECTORID` for connector_id
        6. `CAROL_DOMAIN` for environment
        7. `CAROLUSER` for carol user email
        8. `CAROLPWD` for user password.

    Args:
        domain: Tenant name. e.x., domain.carol.ai.
        app_name: Carol app name.
        auth: Auth Carol object to handle authentication
        connector_id: Connector Id
        port: Port to be used (when running locally it could change)
        verbose: If True will print the header, method and URL of each API call.
        organization: Organization domain.
        environment: Which Carol's environment to use.
            There are three possible values today.

                1. 'carol.ai' for the production environment
                2. 'karol.ai' for the explore environment
                3. 'qarol.ai' for the QA environment
        host: This will overwrite the host used. Today the host is:

                1. if organization is None, host={domain}.{environment}
                2. else host={organization}.{environment}

            See Carol._set_host.
        user: User
        password:User passowrd
        api_key: Carol's Api Key
        org_level: If True, will log-in at organization level.
        dotenv_path: Path to dotenv file, if loading is required.

    Raises:
        MissingInfoCarolException if there is any mandatory parameter missing
    """

    def __init__(
        self,
        domain: T.Optional[str] = None,
        app_name: T.Optional[str] = None,
        auth: T.Optional[T.Union[ApiKeyAuth, PwdAuth]] = None,
        connector_id: T.Optional[str] = None,
        port: int = 443,
        verbose: bool = False,
        organization: T.Optional[str] = None,
        environment: T.Optional[str] = None,
        host: T.Optional[str] = None,
        user: T.Optional[str] = None,
        password: T.Optional[str] = None,
        api_key: T.Optional[str] = None,
        org_level: bool = False,
        dotenv_path: T.Optional[T.Union[str, Path]] = None,
    ):
        if dotenv_path is not None:
            dotenv_path = Path(dotenv_path)
            if not dotenv_path.exists():
                msg = f"Invalid path to dotenv: {dotenv_path}"
                raise exceptions.MissingInfoCarolException(msg)
            load_dotenv(str(dotenv_path))

        auth = _prepare_auth(api_key, auth, connector_id, password, user)

        domain = domain or os.getenv("CAROLTENANT") or None
        if (domain is None) and (not org_level):
            raise exceptions.MissingInfoCarolException("`domain` must be set.")

        if org_level is True:
            domain = None

        app_name = app_name or os.getenv("CAROLAPPNAME", " ")
        environment = environment or os.getenv("CAROL_DOMAIN") or "carol.ai"
        organization = organization or os.getenv("CAROLORGANIZATION") or None

        if (domain is None and not org_level) or app_name is None or auth is None:
            raise exceptions.MissingInfoCarolException(
                "domain, app_name and auth must be specified as parameters, either "
                "in the environment variables CAROLTENANT, CAROLAPPNAME,"
                "  CAROLAPPOAUTH or CAROLUSER+CAROLPWD and  CAROLCONNECTORID"
            )

        if environment is None and os.getenv("ENV_DOMAIN") is not None:
            raise exceptions.DeprecatedEnvVarException("ENV_DOMAIN", "CAROL_DOMAIN")

        if organization is None and os.getenv("API_SUBDOMAIN") is not None:
            raise exceptions.DeprecatedEnvVarException(
                "API_SUBDOMAIN", "CAROLORGANIZATION"
            )

        self._current_user: T.Optional[ResponseType] = None
        self._host_string = host
        self._is_org_level = org_level
        self._tenant: T.Optional[ResponseType] = None
        self._user_agent = f"pyCarol/{__version__}"
        self.app_name: T.Optional[str] = app_name
        self.auth = auth
        self.connector_id = auth.connector_id
        self.domain = domain
        self.environment = environment
        self.host = _set_host(self.environment, domain, host, organization)
        self.org = None
        self.organization = organization
        self.port = port
        self.response: T.Optional[ResponseType] = None
        self.session: T.Optional[requests.Session] = None
        self.verbose = verbose

        self.auth.login(self)

    @property
    def current_user(self) -> ResponseType:
        """Return the current user.

        Returns:
            Carol's API response payload.
        """
        endpoint = (
            "v2/orgUsers/current" if self._is_org_level is True else "v2/users/current"
        )
        if self._current_user is None:
            self._current_user = self.call_api(endpoint)
        return self._current_user

    @property
    def tenant(self) -> T.Optional[ResponseType]:
        """Return the current tenant.

        Returns:
            Carol's API response payload.
        """
        if self._tenant is None and not self._is_org_level:
            self._tenant = Tenant(self).get_tenant_by_domain(self.domain)
        return self._tenant

    def call_api(
        self,
        path: str,
        method: T.Optional[str] = None,
        data=None,
        auth: bool = True,
        params=None,
        content_type: T.Optional[str] = "application/json",
        retries: int = 8,
        session: T.Optional[requests.Session] = None,
        backoff_factor: float = 0.5,
        status_forcelist: T.Tuple[int, ...] = (502, 503, 504, 524),
        downloadable: bool = False,
        method_whitelist: T.FrozenSet[str] = frozenset(
            ["HEAD", "TRACE", "GET", "PUT", "OPTIONS", "DELETE", "POST"]
        ),
        errors: str = "raise",
        extra_headers: T.Optional[T.Dict] = None,
        files: T.Optional[T.Dict] = None,
        prefix_path: str = "/api/",
        **kwds,
    ) -> ResponseType:
        """Handle all the API calls.

        Args:
            path: API URI path. e.x.  v2/staging/schema.
            method: Set of uppercased HTTP method verbs that we should call on.
            data: Same type as :class: `Session.request.data`
                Object to send in the body of the request.
            auth: If API call should be authenticated
            params: in the query string for the :class:`requests.Request`.
            content_type: Content type for the api call
            retries: Number of retries for the API calls
            session: It allows you to persist certain parameters across requests.
            backoff_factor: Backoff factor to apply between  attempts. It will sleep
                for: {backoff factor} * (2 ^ ({retries} - 1)) seconds
            status_forcelist: A set of integer HTTP status codes that we should force a
                retry on. A retry is initiated if the request method is in
                method_whitelist and the response status code is in status_forcelist.
            downloadable: If the request will return a file to download.
            method_whitelist: Set of uppercased HTTP method verbs that we should retry
                on.
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response
            extra_headers: extra headers to be sent.
            files: Used when uploading files to carol. This will be sent to
                :class: `requests.request`
            prefix_path: Prefix path to be used to create the final url
                'https://{self.host}:{self.port}{prefix_path}{path}.
            kwds: `dict` default `None`
                Extra parameters to be sent to :class: `requests.request`

        Returns:
            Dict with API response.
        """
        if session is not None:
            self.session = session

        extra_headers = extra_headers or {}
        url = f"https://{self.host}:{self.port}{prefix_path}{path}"

        if method is None:
            method = "GET" if data is None else "POST"

        met_list = [
            "HEAD",
            "TRACE",
            "GET",
            "PUT",
            "POST",
            "OPTIONS",
            "PATCH",
            "DELETE",
            "CONNECT",
        ]
        assert method in met_list, f"API method must be {met_list}"

        headers = {"accept": "application/json"}
        if auth:
            self.auth.authenticate_request(headers)

        data_json = None
        if method == "GET":
            pass

        elif method in ("POST", "DELETE", "PUT"):
            if content_type is not None:
                headers["content-type"] = content_type

            if content_type == "application/json":
                data_json = data
                data = None

        headers.update(extra_headers)
        headers.update({"User-Agent": self._user_agent})

        __count = 0
        while True:
            self.session = _retry_session(
                retries=retries,
                session=self.session,
                backoff_factor=backoff_factor,
                status_forcelist=status_forcelist,
                method_whitelist=method_whitelist,
            )

            response = self.session.request(
                method=method,
                url=url,
                data=data,
                json=data_json,
                headers=headers,
                params=params,
                files=files,
                **kwds,
            )

            if self.verbose:
                data_ = data_json if data_json is not None else data
                print(f"Calling {method} {url}. Payload: {data_}. Params: {params}")
                print(f"        Headers: {headers}")

            if response.ok or errors == "ignore":
                if downloadable:  # Used when downloading carol app file.
                    return response

                response.encoding = "utf-8"
                self.response = response
                if response.text == "":
                    return {}
                return json.loads(response.text)

            if (response.reason == "Unauthorized") and isinstance(self.auth, PwdAuth):
                if response.json().get("possibleResponsibleField") in [
                    "password",
                    "userLogin",
                ]:
                    raise exceptions.InvalidToken(response.text)
                self.auth.get_access_token()  # It will refresh token if Unauthorized
                __count += 1
                if __count < 5:  # To avoid infinity loops
                    continue

                raise Exception(
                    "Too many retries to refresh token.\n",
                    response.text,
                    response.status_code,
                )
            if response.status_code == 404:
                raise exceptions.CarolApiResponseException(
                    response.text, response.status_code
                )

            raise Exception(response.text, response.status_code)

    def issue_api_key(self, connector_id: T.Optional[str] = None) -> ResponseType:
        """Create an API key for a given connector.

        Args:
            connector_id: Connector ID to be used when creating the APIkey

        Returns:
            Dictionary with the API key.
        """
        if connector_id is None:
            connector_id = self.connector_id
        resp = self.call_api(
            "v2/apiKey/issue",
            data={"connectorId": connector_id},
            content_type="application/x-www-form-urlencoded",
        )
        return resp

    def api_key_details(self, api_key: str, connector_id: str) -> ResponseType:
        """Display information about the API key.

        Args:
            api_key: Carol's api key
            connector_id: Connector Id which API key was created.

        Returns:
            Dictionary with API key information.
        """
        resp = self.call_api(
            "v2/apiKey/details", params={"apiKey": api_key, "connectorId": connector_id}
        )

        return resp

    def api_key_revoke(self, connector_id: str) -> ResponseType:
        """Revoke API key for the given connector_id.

        Args:
            connector_id: Connector Id which API key was created.

        Returns:
            Dictionary with API request response.
        """
        resp = self.call_api(
            "v2/apiKey/revoke",
            method="DELETE",
            content_type="application/x-www-form-urlencoded",
            params={"connectorId": connector_id},
        )

        return resp

    def switch_org_level(self):
        """Switch organization level."""
        if self._is_org_level:
            warnings.warn("already in org level.", UserWarning, stacklevel=3)
            return
        org = self._current_org()
        self.auth.switch_org_context(org["mdmId"])
        self._is_org_level = True

    def switch_environment(
        self,
        env_name: T.Optional[str] = None,
        env_id: T.Optional[str] = None,
        app_name: T.Optional[str] = None,
        org_name: T.Optional[str] = None,
        org_id: T.Optional[str] = None,
    ) -> "Carol":
        """Switch org/environments.

        If the user has access to this environment, it will be "logged in" in this new
        org/environment.

        Args:
            env_name: Environment (tenant) name to switch the context to.
            env_id: Environment (tenant) id to switch the context to.
            app_name: App name in the target environment to switch the context to.
                Only needed with using CDS.
            org_name: The organization name to switch context to. If the same keep it
                `None`.
            org_id: The organization id to switch context to. If the same keep it
                `None`.

        Returns:
            Carol

        .. code:: python
            from pycarol import Carol, Staging
            carol = Carol('B', 'teste', auth=PwdAuth('email@totvs.com.br', 'pass'), )
            carol.switch_environment('A')
            Staging(carol_tenant_A).fetch_parquet(...) # fetch parquet from tenant A
            # To switch back
            carol.switch_environment('B')
            #back to tenant B
        """
        if self.org is None:
            self.org = Organization(self).get_organization_info(self.organization)

        current = self.get_current()

        if (org_id is not None and org_id != current["org_id"]) or (
            org_name is not None and org_name != current["org_name"]
        ):
            # Switch to org context.
            if not current["org_level"]:
                self.switch_org_level()
            if org_id is None:
                org_id = Organization(self).get_organization_info(org_name)["mdmId"]
            # Switch org.
            self.auth.switch_org_context(org_id)

        if env_name:
            env_id = Tenant(self).get_tenant_by_domain(env_name)["mdmId"]
        if env_id is not None:
            self.auth.switch_context(env_id=env_id)
            self._tenant = self._current_env()
            self._is_org_level = False
        else:
            self._tenant = None
            self._is_org_level = True

        self.domain = env_name
        # TODO: Today we cannot use CDS without a valid app name.
        self.app_name = app_name
        self.organization = self._current_org()["mdmName"]

        self.host = _set_host(
            domain=self.domain,
            organization=self.organization,
            environment=self.environment,
            host=self._host_string,
        )

        return self

    def switch_context(
        self,
        env_name: T.Optional[str] = None,
        env_id: T.Optional[str] = None,
        app_name: T.Optional[str] = None,
        org_name: T.Optional[str] = None,
        org_id: T.Optional[str] = None,
    ) -> SwitchContext:
        """Context manager to temporary have access to a second environment.

        Args:
            env_name: Environment (tenant) name to switch the context to.
            env_id: Environment (tenant) id to switch the context to.
            app_name: App name in the target environment to switch the context to.
                Only needed with using CDS.

        Returns:
            SwitchContext

        Examples:
            .. code:: python

                from pycarol import Carol, Staging
                carol = Carol('B', 'teste', auth=PwdAuth('email@totvs.com.br', 'pwd'), )
                with carol.switch_context('A') as carol_tenant_A:
                    # fetch parquet from tenant A
                    Staging(carol_tenant_A).fetch_parquet(...)
                #back to tenant B
        """
        if self.org is None:
            self.org = Organization(self).get_organization_info(self.organization)

        if env_name:
            env_id = Tenant(self).get_tenant_by_domain(env_name)["mdmId"]

        return SwitchContext(
            parent_context=copy.deepcopy(self),
            env_name=env_name,
            env_id=env_id,
            app_name=app_name,
            org_name=org_name,
            org_id=org_id,
        )

    def _current_env(self):
        return self.call_api("v1/tenants/current", errors="ignore")

    def _current_org(self):
        return self.call_api("v1/organizations/current")

    def get_current(self, level: str = "all") -> T.Dict[str, T.Any]:
        """Get current org/env information.

        Args:
            level: Possible Values:
                "all": To get organization and environment information.
                "org": To get organization information.
                "env": To get environment information.

        Returns:
            Dictionary with keys org_Id, org_name, env_id, env_name

        Raises:
            ValueError when level not in 'org', 'env' or 'all'
        """
        env = {}
        org = {}

        if level.lower() not in ["org", "env", "all"]:
            raise ValueError(
                f"level should be 'all', 'org', 'env', {level} was passed."
            )

        if level.lower() in ("env", "all"):
            env = self._current_env()

        if level.lower() in ("org", "all"):
            org = self._current_org()

        self._is_org_level = env.get("mdmName") is None

        return {
            "env_name": env.get("mdmName"),
            "env_id": env.get("mdmId"),
            "org_name": org.get("mdmName"),
            "org_id": org.get("mdmId"),
            "org_level": self._is_org_level,
        }

    def get_tenants_for_user(self) -> ResponseType:
        """Get all tenants for the current user.

        Returns:
            Dict
        """
        return self.call_api("v1/users/assignedTenantsForCurrentUser", method="GET")


def _prepare_auth(
    app_oauth: T.Optional[str] = None,
    auth: T.Optional[T.Union[ApiKeyAuth, PwdAuth]] = None,
    connector_id: T.Optional[str] = None,
    password: T.Optional[str] = None,
    username: T.Optional[str] = None,
) -> T.Union[ApiKeyAuth, PwdAuth]:
    app_oauth = app_oauth or os.getenv("CAROLAPPOAUTH") or None
    connector_id = (
        connector_id or os.getenv("CAROLCONNECTORID") or __CONNECTOR_PYCAROL__
    )
    password = password or os.getenv("CAROLPWD") or None
    username = username or os.getenv("CAROLUSER") or None

    if auth is not None:
        if auth.connector_id is None:
            auth.set_connector_id(connector_id)
        return auth

    if username is not None and password is not None:
        auth = PwdAuth(user=username, password=password)
        auth.set_connector_id(connector_id)
        return auth

    if app_oauth is not None and connector_id is not None:
        auth = ApiKeyAuth(app_oauth)
        auth.set_connector_id(connector_id)
        return auth

    raise exceptions.MissingInfoCarolException(
        "either `auth` or `username/password` or `api_key` or "
        "pycarol env variables must be set."
    )


def _set_host(
    environment: str,
    domain: T.Optional[str],
    host: T.Optional[str] = None,
    organization: T.Optional[str] = None,
) -> str:
    """Set the host to be used.

    Args:
        domain: Tenant name.
            e.x., tenant.carol.ai
        environment: There are three possible values today:
            1. 'carol.ai' for the production environment
            2. 'karol.ai' for the explore environment
            3. 'qarol.ai' for the QA environment
        host: Host to be used.
        organization: Organization domain.

    Returns:
        host
    """
    if host is not None:
        return host
    if organization is not None:
        return f"{organization}.{environment}"
    return f"{domain}.{environment}"


def _retry_session(
    retries: int = 5,
    session: T.Optional[requests.Session] = None,
    backoff_factor: float = 0.5,
    status_forcelist: T.Tuple[int, ...] = (500, 502, 503, 504, 524),
    method_whitelist: T.FrozenSet[str] = frozenset(
        ["HEAD", "TRACE", "GET", "PUT", "OPTIONS", "DELETE"]
    ),
) -> requests.Session:
    """Handle retries between calls.

    Args:
        retries: Number of retries for the API calls
        session: Session object dealt `None`
            It allows you to persist certain parameters across requests.
        backoff_factor: Backoff factor to apply between  attempts. It will sleep
            for: {backoff factor} * (2 ^ ({retries} - 1)) seconds
        status_forcelist: A set of integer HTTP status codes that we should force a
            retry on. A retry is initiated if the request method is in
            method_whitelist and the response status code is in status_forcelist.
        method_whitelist: `iterable` , default frozenset(['HEAD', 'TRACE', 'GET',
            'PUT', 'OPTIONS', 'DELETE'])). Set of uppercased HTTP method verbs that
            we should retry on.

    Returns:
        session
    """
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        method_whitelist=method_whitelist,
    )
    adapter = requests.adapters.HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



class CarolApiResponseException(Exception):

    """Custom Exception to handle Exception on Carol API Calls."""

    pass


class InvalidToken(Exception):

    """Custom Exception to handle Invalid token."""

    pass


class MissingInfoCarolException(ValueError):

    """Custom exception to handle Carol missing information on Carol.__init__."""

    pass


class DeprecatedEnvVarException(Exception):

    """Custom exception to handle deprecated environment variables.

    Args:
        old_var: Deprecated environment variable.
        new_var: Environment variable to be replaced.
    """

    def __init__(self, old_var: str, new_var: str):
        msg = f"""
            {old_var} environment var is deprecated.
            Please replace it by {new_var}.
        """
        super().__init__(msg)


class NoServiceAccountException(Exception):

    """Custom exception to handle missing BQ service account."""

    pass


class NotListAsCallResponseException(Exception):

    """Custom exception to handle expected type as list."""

    def __init__(self):
        msg = "Expected type for API response must be a list."
        super().__init__(msg)


class NotMapAsCallResponseException(Exception):

    """Custom exception to handle expected type as dict."""

    def __init__(self):
        msg = "Expected type for API response must be a dict."
        super().__init__(msg)


class NotResponseAsCallResponseException(Exception):

    """Custom exception to handle expected type as requests.Response."""

    def __init__(self):
        msg = "Expected type for API response must be a requests.Response."
        super().__init__(msg)


class RepeatedMDMIdsException(Exception):

    """Custom exception for query error when receiving multiple times same mdmId."""

    def __init__(self):
        msg = "Repeated mdmId's. Something is wrong"
        super().__init__(msg)


class NoScrollIdException(Exception):

    """Custom exception for query error when no scrollId is present.."""

    def __init__(self):
        msg = "No scrollId. Something is wrong"
        super().__init__(msg)


class PandasNotFoundException(Exception):

    """Custom exception for query error when no scrollId is present.."""

    def __init__(self):
        msg = (
            "You are trying to get a pandas DataFrame (return_dataframe is True)."
            " However, pandas is not installed."
            " Please install pandas ('pip install pandas') or pass return_dataframe as"
            " False."
        )
        super().__init__(msg)



import logging
from collections import defaultdict
import random
import time
from itertools import chain

from pycarol import (
    Carol, ApiKeyAuth, PwdAuth, Tasks, Staging, Connectors, CDSStaging, Subscription, DataModel, Apps, CDSGolden
)
from pycarol.query import delete_golden

def track_tasks(carol, task_list, retry_count=3, logger=None, callback=None, polling_delay=5):
    """Track a list of taks from carol, waiting for errors/completeness. 

    Args:

        carol (pycarol.Carol): pycarol.Carol instance
        task_list (list): List of tasks in Carol
        retry_count (int, optional): Number of times to restart a failed task. Defaults to 3.
        logger (logging.logger, optional): logger to log information. Defaults to None.
        callback (calable, optional): This function will be called every time task status are fetched from carol. 
            A dictionary with task status will be passed to the function. Defaults to None.
        polling_delay (int, optional): Time in seconds to pull task status from Carol

    Usage:

    .. code:: python

        from pycarol import Carol
        from pycarol.functions import track_tasks
        carol = Carol()
        def callback(task_list):
            print(task_list)
        track_tasks(carol=carol, task_list=['task_id_1', 'task_id_2'], callback=callback)

    Returns:

        [dict, bool]: dict with status of each task and booling if any task failed more than retry_count times.
    
    """
    
    if logger is None:
        logger = logging.getLogger(carol.domain)

    retry_tasks = defaultdict(int)
    n_task = len(task_list)
    max_retries = set()
    carol_task = Tasks(carol)
    while True:
        task_status = defaultdict(list)
        for task in task_list:
            status = carol_task.get_task(task).task_status
            task_status[status].append(task)
        for task in task_status['FAILED'] + task_status['CANCELED']:
            logger.warning(f'Something went wrong while processing: {task}')
            retry_tasks[task] += 1
            if retry_tasks[task] > retry_count:
                max_retries.update([task])
                logger.error(
                    f'Task: {task} failed {retry_count} times. will not restart')
                continue

            logger.info(f'Retry task: {task}')
            carol_task.reprocess(task)

        if len(task_status['COMPLETED']) == n_task:
            logger.debug(f'All task finished')
            return task_status, False

        elif len(max_retries) + len(task_status['COMPLETED']) == n_task:
            logger.warning(f'There are {len(max_retries)} failed tasks.')
            return task_status, True
        else:
            time.sleep(round(polling_delay + random.random(), 2))
            logger.debug('Waiting for tasks')
        if callable(callback):
            callback(task_status)


def delele_all_golden_data(carol, dm_name):
    """Delete golden files from a datamodel in all storages.

    Args:
        carol (pycarol.Carol): Carol instance
        dm_name (str): Data Model name

    Returns:
        list: list of tasks created
    """
    
    cds_CDSGolden = CDSGolden(carol)
    t = []
    dm_id = DataModel(carol).get_by_name(dm_name)['mdmId']
    task = cds_CDSGolden.delete_rejected(dm_id=dm_id, )

    t += [task]
    task = cds_CDSGolden.delete(dm_id=dm_id, )
    delete_golden(carol, dm_name)
    t += [task['taskId'], ]
    return t


def par_delete_golden(carol, dm_list, n_jobs=5):
    """
    Deletes golden files from a list of datamodels in parallel.

    Args:
        carol (pycarol.Carol): Carol instance
        dm_list (list): List of datamodels
        n_jobs (int, optional): Number of parallel jobs. Defaults to 5.

    Returns:
        list: list of tasks created
    """
    from joblib import Parallel, delayed
    tasks = Parallel(n_jobs=n_jobs)(delayed(delele_all_golden_data)(carol, i)
                                    for i in dm_list)
    return list(chain(*tasks))


def delete_staging_data(carol, staging_name, connector_name):
    """Delete a staging.

    Args:
        carol (pycarol.Carol): Login instance
        staging_name (str): Staging name
        connector_name (str): Connector name
    """ 

    t = []
    cds_ = CDSStaging(carol)
    task = cds_.delete(staging_name=staging_name,
                        connector_name=connector_name)
    t += [task['taskId'], ]
    return t


def par_delete_staging(carol, staging_list, connector_name, n_jobs=5):
    """
    Deletes staging files from a list of datamodels in parallel.

    Args:
        carol (pycarol.Carol): Login instance
        staging_list (list): List of datamodels
        connector_name (str): Connector name
        n_jobs (int, optional): Number of parallel jobs. Defaults to 5.

    Returns:
        list: list of tasks created
    """
    from joblib import Parallel, delayed
    tasks = Parallel(n_jobs=n_jobs)(delayed(delete_staging_data)(carol, i, connector_name)
                                    for i in staging_list)
    return list(chain(*tasks))



"""Back-end for BigQuery-related code."""
import copy
from datetime import datetime
import json
from pathlib import Path
import sys
import typing as T

from google.cloud import bigquery, bigquery_storage, bigquery_storage_v1
from google.cloud.bigquery_storage import types
from google.oauth2.service_account import Credentials

try:
    import pandas
except ImportError:
    pass

from . import __TEMP_STORAGE__
from . import exceptions
from .carol import Carol
from .storage import Storage


class Token:

    """Stores Token/Service Account information.

    Args:
        service_account: service account.
        env: dict from Carol.get_current().

    Attributes:
        expiration_time: datetime of service account expiration.
        env: dict from Carol.get_current().
        service_account: provided by Carol.
    """

    def __init__(
        self,
        service_account: T.Dict[str, T.Any],
        env: T.Dict[str, str],
    ):
        self.expiration_time = service_account["expiration_time"]
        self._env = env
        self.service_account = service_account

    def to_dict(self) -> T.Dict[str, T.Any]:
        """Convert Token to dictionary.

        Returns: dict Token.
        """
        return {
            "service_account": self.service_account,
            "env": copy.deepcopy(self._env),
            "expiration_time": self.expiration_time,
        }

    def expired(self) -> bool:
        """Check if token has expired.

        Return True if has expired.
        """
        dt_format = "%Y-%m-%dT%H:%M:%S.%fZ"
        expiration_time_ = datetime.strptime(self.expiration_time, dt_format)
        return expiration_time_ < datetime.utcnow()


class TokenManager:

    """Manages Google service accounts to allow BigQuery operations.

    Args:
        carol: Carol object.
        provided_sa: if not provided, it will automatically generate.
        cache_cds: either cache service account to google cloud or not so that other
                   uses may share it. This may be helpful to reduce the number of
                   generated service accounts.
        expiration_window: Number of hours for credentials to expire. Max value 24.

    Attributes:
        expiration_window: Number of hours for credentials to expire. Max value 24.
        cache_cds: either cache service account to google cloud or not so that other
                   uses may share it. This may be helpful to reduce the number of
                   generated service accounts.
        token: Token object containing service account and metadata.
    """

    def __init__(
        self,
        carol: Carol,
        provided_sa: T.Optional[T.Dict[str, T.Any]] = None,
        cache_cds: bool = True,
        expiration_window: int = 24,
    ):
        self._carol = carol
        self.expiration_window = expiration_window
        self._env = carol.get_current()
        temp_file_name = f".pycarol_temp_{self._env['env_id']}.json"
        self._tmp_filepath = Path(__TEMP_STORAGE__) / temp_file_name
        self.cache_cds = cache_cds
        self._storage = Storage(self._carol)

        self.token: T.Optional[Token] = None
        if provided_sa is not None:
            self.token = Token(provided_sa, self._env)
        else:
            self.token = self.get_token()

    def _issue_new_key(self) -> T.Dict[str, T.Any]:
        url = "v1/create_temporary_key"
        prefix_path = "/sql/v1/api/"
        env = self._carol.get_current()
        payload = {
            "expirationTime": self.expiration_window,
            "mdmOrgId": env["org_id"],
            "mdmTenantId": env["env_id"],
        }
        call_ret = self._carol.call_api(
            method="POST", path=url, prefix_path=prefix_path, data=payload
        )
        if call_ret is None:
            msg = "No response from service account."
            raise exceptions.NoServiceAccountException(msg)

        if not isinstance(call_ret, dict):
            msg = "Response from call must be a dict."
            raise exceptions.NoServiceAccountException(msg)

        return call_ret

    def _save_token_file(self, token: Token) -> None:
        self._tmp_filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(self._tmp_filepath, "w", encoding="utf-8") as file:
            json.dump(token.to_dict(), file)

    def _save_token_cloud(self) -> None:
        if not self._tmp_filepath.exists():
            return

        self._storage.save(
            name=self._tmp_filepath.name,
            obj=str(self._tmp_filepath),
            format="file",
            storage_space="pycarol",
        )

    def _load_token_file(self) -> T.Optional[Token]:
        if not self._tmp_filepath.exists():
            return None
        with open(self._tmp_filepath, "r", encoding="utf-8") as file:
            token_dict = json.load(file)

        if "service_account" not in token_dict or "env" not in token_dict:
            return None

        token = Token(token_dict["service_account"], token_dict["env"])
        return token

    def _load_token_cloud(self) -> T.Optional[Token]:
        filename = self._tmp_filepath.name
        cache_exists = self._storage.exists(name=filename, storage_space="pycarol")
        if cache_exists is True:
            sa_path = self._storage.load(
                name=filename,
                format="file",
                storage_space="pycarol",
                cache=False,
            )
            Path(sa_path).rename(self._tmp_filepath)
            return self._load_token_file()
        return None

    def get_forced_token(self) -> Token:
        """Get new service account for BigQuery and cache it.

        Returns:
            Service account

        .. code:: python


            from pycarol import Carol
            from pycarol.bigquery import TokenManager

            tm = TokenManager(Carol())
            service_account = tm.get_forced_token().service_account
        """
        service_account = self._issue_new_key()
        self.token = Token(service_account, self._env)
        self._save_token_file(self.token)
        if self.cache_cds is True:
            self._save_token_cloud()
        return self.token

    def get_token(self) -> Token:
        """Get service account for BigQuery and cache it.

        Returns:
            Service account

        .. code:: python


            from pycarol import Carol
            from pycarol.bigquery import TokenManager

            tm = TokenManager(Carol())
            service_account = tm.get_token().service_account
        """
        if self.token is not None and not self.token.expired():
            return self.token

        local_token = self._load_token_file()
        if local_token is not None and not local_token.expired():
            self.token = local_token
            return self.token

        cloud_token = self._load_token_cloud()
        if cloud_token is not None and not cloud_token.expired():
            self.token = cloud_token
            return self.token

        self.token = self.get_forced_token()
        return self.token


class BQ:

    """Handles BigQuery queries.

    Args:
        carol: object from Carol class.
        service_account: Google Cloud SA with access to BQ.
        cache_cds: if SA should be cached for subsequent uses.
    """

    def __init__(
        self,
        carol: Carol,
        service_account: T.Optional[T.Dict[str, T.Any]] = None,
        cache_cds: bool = True,
    ):
        self._env = carol.get_current()
        self._project_id = f"carol-{self._env['env_id'][0:20]}"
        self._dataset_id = f"{self._project_id}.{self._env['env_id']}"
        self._token_manager = TokenManager(carol, service_account, cache_cds)

    @staticmethod
    def _generate_client(service_account: T.Dict) -> bigquery.Client:
        credentials = Credentials.from_service_account_info(service_account)
        project = service_account["project_id"]
        client = bigquery.Client(project=project, credentials=credentials)
        return client

    def query(
        self,
        query: str,
        dataset_id: T.Optional[str] = None,
        return_dataframe: bool = True,
    ) -> T.Union["pandas.DataFrame", T.List[T.Dict[str, T.Any]]]:
        """Run query. This will generate a SA if necessary.

        Args:
            query: BigQuery SQL query.
            dataset_id: BigQuery dataset ID, if not provided, it will use the default
                        one.
            return_dataframe: Return dataframe if True.

        Returns:
            Query result.

        Usage:

        .. code:: python


            from pycarol import BQ, Carol

            bq = BQ(Carol())
            query = 'select * from invoice limit 10'
            df = bq.query(query, return_dataframe=True)
        """
        service_account = self._token_manager.get_token().service_account
        client = self._generate_client(service_account)

        dataset_id = dataset_id or self._dataset_id
        job_config = bigquery.QueryJobConfig(default_dataset=dataset_id)
        results_job = client.query(query, job_config=job_config)

        results = [dict(row) for row in results_job]

        if return_dataframe is False:
            return results

        if "pandas" not in sys.modules and return_dataframe is True:
            raise exceptions.PandasNotFoundException

        return pandas.DataFrame(results)


class BQStorage:

    """Handles BigQuery Storage API queries.

    Args:
        carol: object from Carol class.
        service_account: Google Cloud SA with access to BQ.
        cache_cds: if SA should be cached for subsequent uses.
    """

    def __init__(
        self,
        carol: Carol,
        service_account: T.Optional[T.Dict[str, T.Any]] = None,
        cache_cds: bool = True,
    ):
        self._env = carol.get_current()
        self._project_id = f"carol-{self._env['env_id'][0:20]}"
        self._dataset_id = f"{self._env['env_id']}"
        self._token_manager = TokenManager(carol, service_account, cache_cds)

    @staticmethod
    def _generate_client(service_account) -> bigquery_storage.BigQueryReadClient:
        credentials = Credentials.from_service_account_info(service_account)
        return bigquery_storage.BigQueryReadClient(credentials=credentials)

    def _get_read_session(
        self,
        client: bigquery_storage.BigQueryReadClient,
        table_name: str,
        columns_names: T.Optional[T.List[str]] = None,
    ) -> bigquery_storage_v1.types.ReadSession:
        read_options = None
        if columns_names is not None:
            read_options = types.ReadSession.TableReadOptions(  # type:ignore # noqa:E501 pylint:disable=no-member
                selected_fields=columns_names
            )

        table_path = f"projects/{self._project_id}/datasets/{self._dataset_id}/tables/{table_name}"  # noqa:E501
        requested_session = types.ReadSession(  # type:ignore # noqa:E501 pylint:disable=no-member
            table=table_path,
            data_format=types.DataFormat.ARROW,  # type:ignore # noqa:E501 pylint:disable=no-member
            read_options=read_options,
        )
        parent = f"projects/{self._project_id}"
        read_session = client.create_read_session(
            parent=parent,
            read_session=requested_session,
            max_stream_count=4,
        )
        return read_session

    def query(
        self,
        table_name: str,
        columns_names: T.Optional[T.List[str]] = None,
        return_dataframe: bool = True,
    ) -> T.Union["pandas.DataFrame", T.List[bigquery_storage_v1.reader.ReadRowsPage]]:
        """Read from BigQuery Storage API.

        Args:
            table_name: name of the table (views are not supported).
            columns_names: names of columns to return.
            return_dataframe: if True, return a pandas DataFrame.

        Returns:
            Query result.

        Usage:

        .. code:: python


            from pycarol import BQStorage, Carol

            bq = BQStorage(Carol())
            table_name = "ingestion_stg_model_deep_audit"
            col_names = ["request_id", "version"]
            df = bq.query(table_name, col_names, return_dataframe=True)
        """
        service_account = self._token_manager.get_token().service_account
        client = self._generate_client(service_account)
        read_session = self._get_read_session(client, table_name, columns_names)

        stream = read_session.streams[0]
        reader = client.read_rows(stream.name)

        frames = []
        for frame in reader.rows().pages:
            frames.append(frame)

        if return_dataframe is False:
            return frames

        if "pandas" not in sys.modules and return_dataframe is True:
            raise exceptions.PandasNotFoundException

        dataframe = pandas.concat([frame.to_dataframe() for frame in frames])
        dataframe = dataframe.reset_index(drop=True)
        return dataframe



from unittest import mock

import pycarol


def test__init() -> None:
    """Apps() must initialize carol attribute."""
    carol = mock.MagicMock()
    ret = pycarol.Apps(carol)

    assert ret.carol == carol


def test_all() -> None:
    """apps.Apps.all should return whatever /v1/tenantApps responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {"hits": [{"mdmName": "test"}], "totalHits": 54}
    mock_apps.carol.call_api.return_value = call_api_ret
    ret = pycarol.Apps.all(mock_apps)
    assert ret == {"test": {"mdmName": "test"}}
    assert mock_apps.carol.call_api.mock_calls[0][1] == ("v1/tenantApps",)


def test_get_by_name() -> None:
    """apps.Apps.get_by_name should return whatever /v1/tenantApps/name responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    ret = pycarol.Apps.get_by_name(mock_apps, app_name)
    assert ret == call_api_ret
    url = f"v1/tenantApps/name/{app_name}"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_get_by_id() -> None:
    """apps.Apps.get_by_id should return whatever /v1/tenantApps responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_id = "test"
    ret = pycarol.Apps.get_by_id(mock_apps, app_id)
    assert ret == call_api_ret
    url = f"v1/tenantApps/{app_id}"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_get_settings() -> None:
    """apps.Apps.get_settings should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {
        "mdmTenantAppSettingValues": [{"mdmName": "set1", "mdmParameterValue": "val1"}]
    }
    mock_apps.carol.call_api.return_value = call_api_ret
    app_id = "test"
    ret = pycarol.Apps.get_settings(mock_apps, app_id=app_id)
    assert ret == {"set1": "val1"}
    url = f"v1/tenantApps/{mock_apps.current_app_id}/settings"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_get_manifest() -> None:
    """apps.Apps.get_manifest should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    ret = pycarol.Apps.get_manifest(mock_apps, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/tenantApps/manifest/{app_name}"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_edit_manifest() -> None:
    """apps.Apps.edit_manifest should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    ret = pycarol.Apps.edit_manifest(mock_apps, {}, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/tenantApps/manifest/{app_name}"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_get_git_process() -> None:
    """apps.Apps.get_git_process should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = [{"test": "test"}]
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    ret = pycarol.Apps.get_git_process(mock_apps, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/compute/{app_name}/getProcessesByGitRepo"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_build_docker_git() -> None:
    """apps.Apps.build_docker_git should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    manifest = [
        {"dockerName": "name", "dockerTag": "tag", "instanceType": "type"},
        {"dockerName": "name", "dockerTag": "tag", "instanceType": "type"},
    ]
    mock_apps.get_git_process = lambda app_name: manifest
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    ret = pycarol.Apps.build_docker_git(mock_apps, "toke", app_name=app_name)
    assert ret == [call_api_ret, call_api_ret]
    url = f"v1/compute/{app_name}/buildGitDocker"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)
    assert mock_apps.carol.call_api.mock_calls[1][1] == (url,)


def test_update_setting_values() -> None:
    """apps.Apps.update_setting_values should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    app_id = "id"
    mock_apps.get_by_name = lambda x: {"mdmId": app_id}
    settings_id = "settings_id"
    mock_apps._get_app_settings_config = lambda app_id: {"mdmId": settings_id}
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    ret = pycarol.Apps.update_setting_values(mock_apps, {}, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/tenantApps/{app_id}/settings/{settings_id}?publish=true"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_start_app_process() -> None:
    """apps.Apps.start_app_process should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    app_id = "id"
    mock_apps.get_by_name = lambda x: {"mdmId": app_id}
    process_id = "process_id"
    mock_apps.get_processes_info = lambda app_name: {"mdmId": process_id}
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    process_name = "process_name"
    ret = pycarol.Apps.start_app_process(mock_apps, process_name, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/tenantApps/{app_id}/aiprocesses/{process_id}/execute/{process_name}"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_get_processes_info() -> None:
    """apps.Apps.get_processes_info should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    app_id = "id"
    mock_apps.get_by_name = lambda x: {"mdmId": app_id}
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    ret = pycarol.Apps.get_processes_info(mock_apps, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/tenantApps/{app_id}/aiprocesses"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_get_subscribable_carol_apps() -> None:
    """apps.Apps.get_subscribable_carol_apps should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {"hits": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    ret = pycarol.Apps.get_subscribable_carol_apps(mock_apps)
    assert ret == "test"
    url = "v1/tenantApps/subscribableCarolApps"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_install_carol_app() -> None:
    """apps.Apps.install_carol_app should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    app_id = "id"
    call_api_ret = {"mdmId": app_id}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    app_version = "version"

    mock_apps.get_subscribable_carol_apps = lambda: [
        {"mdmName": app_name, "mdmId": app_id, "mdmAppVersion": app_version}
    ]
    ret = pycarol.Apps.install_carol_app(
        mock_apps, app_name=app_name, app_version=app_version
    )
    assert ret == call_api_ret
    url1 = f"v1/tenantApps/subscribe/carolApps/{app_id}"
    url2 = "v1/tenantApps/id/install"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url1,)
    assert mock_apps.carol.call_api.mock_calls[1][1] == (url2,)


def test_get_app_details() -> None:
    """apps.Apps.get_app_details should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    app_id = "id"
    mock_apps.get_subscribable_carol_apps = lambda: [
        {"mdmName": app_name, "mdmId": app_id}
    ]
    ret = pycarol.Apps.get_app_details(mock_apps, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/carolApps/{app_id}/details"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)


def test_upload_file() -> None:
    """apps.Apps.upload_file should return whatever call_api responds."""
    mock_apps = mock.MagicMock()
    app_id = "id"
    mock_apps.get_by_name = lambda x: {"mdmCarolAppId": app_id}
    call_api_ret = {"test": "test"}
    mock_apps.carol.call_api.return_value = call_api_ret
    app_name = "test"
    filepath = "/tmp/file.zip"
    with open(filepath, "w", encoding="utf-8") as file_:
        file_.write("test")

    ret = pycarol.Apps.upload_file(mock_apps, filepath, app_name=app_name)
    assert ret == call_api_ret
    url = f"v1/carolApps/{app_id}/files/upload"
    assert mock_apps.carol.call_api.mock_calls[0][1] == (url,)



from pycarol import Carol, Staging, Connectors, DataModel
from collections import defaultdict
import itertools


def find_keys(node, kv):
    """Find recursively all the values from a given key. 

    Args:
        node (dict of dict): Nested dictionary.
        kv (str): dictionary key to find. 

    Yields:
        dict: dictionary
    """
    if isinstance(node, list):
        for i in node:
            yield from find_keys(i, kv)
    elif isinstance(node, dict):
        if kv in node:
            yield node[kv]
        for j in node.values():
            yield from find_keys(j, kv)


def unroll_list(l):
    """Unroll a list of lists to a flat list

    Args:
        l (list of lists): List of list to unroll

    Yields:
        list: unrolled list.
    """

    if isinstance(l, list):
        for i in l:
            yield from unroll_list(i)
    else:
        yield l


def get_staging_prefix(carol, staging_prefix, connector_name, connector_id):
    """Create the prefix for a given set of staging_prefix, connector_name, connector_id

    Args:
        carol: `pycarol.Carol`
            Carol() object.
        connector_name: `str`
            Connector Name
        prefix: 'str` default `DM_`
            prefix to add to the data model name. e.g., if dm_name='mydatamoldel', the result will be "DM_mydatamoldel`.
            This is only applied for DataModel.
        staging_prefix: 'str` default `None`
            Prefix for the staging name. e.g., if staging_prefix='connector_name` the output will be:
                    { "connector_name_stag3" : {"connector_name_stag1", "connector_name_stag2"}}
            Possible values are: 'connector_name', 'connector_id', None

    """
    conn = Connectors(carol)
    if staging_prefix is None:
        staging_prefix = ''
    elif staging_prefix == 'connector_name':
        if connector_name is None:
            connector_name = conn.get_by_id(connector_id)['mdmName']
        staging_prefix = connector_name + '_'
    elif staging_prefix == 'connector_id':
        if connector_id is None:
            connector_id = conn.get_by_name(connector_name)['mdmId']
        staging_prefix = connector_id + '_'
    else:
        raise ValueError(
            f'`staging_prefix` should be either `connector_name`, `connector_id` or `None` {staging_prefix} was used'
        )
    return staging_prefix, connector_name, connector_id
